# Copyright (c) OpenMMLab. All rights reserved.
from .interclr_memory import InterCLRMemory
from .odc_memory import ODCMemory
from .simple_memory import SimpleMemory

__all__ = ['InterCLRMemory', 'ODCMemory', 'SimpleMemory']



_base_ = [
    '../_base_/models/simsiam.py',
    '../_base_/datasets/imagenet_mocov2.py',
    '../_base_/schedules/sgd_coslr-200e_in1k.py',
    '../_base_/default_runtime.py',
]

# set base learning rate
lr = 0.05

# additional hooks
custom_hooks = [
    dict(type='SimSiamHook', priority='HIGH', fix_pred_lr=True, lr=lr)
]

# optimizer
optimizer = dict(lr=lr, paramwise_options={'predictor': dict(fix_lr=True)})

# runtime settings
# the max_keep_ckpts controls the max number of ckpt file in your work_dirs
# if it is 3, when CheckpointHook (in mmcv) saves the 4th ckpt
# it will remove the oldest one to keep the number of total ckpts as 3
checkpoint_config = dict(interval=10, max_keep_ckpts=3)



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np
import torch
import torch.distributed as dist
from mmcv.runner import HOOKS, Hook
from mmcv.utils import print_log

from mmselfsup.utils import Extractor
from mmselfsup.utils import clustering as _clustering


@HOOKS.register_module()
class InterCLRHook(Hook):
    """Hook for InterCLR.

    This hook includes the clustering process in InterCLR.

    Args:
        extractor (dict): Config dict for feature extraction.
        clustering (dict): Config dict that specifies the clustering algorithm.
        centroids_update_interval (int): Frequency of iterations to
            update centroids.
        deal_with_small_clusters_interval (int): Frequency of iterations to
            deal with small clusters.
        evaluate_interval (int): Frequency of iterations to evaluate clusters.
        warmup_epochs (int, optional): The number of warmup epochs to set
            ``intra_loss_weight=1`` and ``inter_loss_weight=0``. Defaults to 0.
        init_memory (bool): Whether to initialize memory banks used in online
            labels. Defaults to True.
        initial (bool): Whether to call the hook initially. Defaults to True.
        online_labels (bool): Whether to use online labels. Defaults to True.
        interval (int): Frequency of epochs to call the hook. Defaults to 1.
        dist_mode (bool): Use distributed training or not. Defaults to True.
        data_loaders (DataLoader): A PyTorch dataloader. Defaults to None.
    """

    def __init__(
            self,
            extractor,
            clustering,
            centroids_update_interval,
            deal_with_small_clusters_interval,
            evaluate_interval,
            warmup_epochs=0,
            init_memory=True,
            initial=True,
            online_labels=True,
            interval=1,  # same as the checkpoint interval
            dist_mode=True,
            data_loaders=None):
        assert dist_mode, 'non-dist mode is not implemented'
        self.extractor = Extractor(dist_mode=dist_mode, **extractor)
        self.clustering_type = clustering.pop('type')
        self.clustering_cfg = clustering
        self.centroids_update_interval = centroids_update_interval
        self.deal_with_small_clusters_interval = \
            deal_with_small_clusters_interval
        self.evaluate_interval = evaluate_interval
        self.warmup_epochs = warmup_epochs
        self.init_memory = init_memory
        self.initial = initial
        self.online_labels = online_labels
        self.interval = interval
        self.dist_mode = dist_mode
        self.data_loaders = data_loaders

    def before_run(self, runner):
        assert hasattr(runner.model.module, 'intra_loss_weight'), \
            "The runner must have attribute \"intra_loss_weight\" in InterCLR."
        assert hasattr(runner.model.module, 'inter_loss_weight'), \
            "The runner must have attribute \"inter_loss_weight\" in InterCLR."
        self.intra_loss_weight = runner.model.module.intra_loss_weight
        self.inter_loss_weight = runner.model.module.inter_loss_weight
        if self.initial:
            if runner.epoch > 0 and self.online_labels:
                if runner.rank == 0:
                    print(f'Resuming memory banks from epoch {runner.epoch}')
                    features = np.load(
                        f'{runner.work_dir}/feature_epoch_{runner.epoch}.npy')
                else:
                    features = None
                loaded_labels = np.load(
                    f'{runner.work_dir}/cluster_epoch_{runner.epoch}.npy')
                runner.model.module.memory_bank.init_memory(
                    features, loaded_labels)
                return

            self.deepcluster(runner)

    def before_train_epoch(self, runner):
        cur_epoch = runner.epoch
        if cur_epoch >= self.warmup_epochs:
            runner.model.module.intra_loss_weight = self.intra_loss_weight
            runner.model.module.inter_loss_weight = self.inter_loss_weight
        else:
            runner.model.module.intra_loss_weight = 1.
            runner.model.module.inter_loss_weight = 0.

    def after_train_iter(self, runner):
        if not self.online_labels:
            return
        # centroids update
        if self.every_n_iters(runner, self.centroids_update_interval):
            runner.model.module.memory_bank.update_centroids_memory()

        # deal with small clusters
        if self.every_n_iters(runner, self.deal_with_small_clusters_interval):
            runner.model.module.memory_bank.deal_with_small_clusters()

        # evaluate
        if self.every_n_iters(runner, self.evaluate_interval):
            new_labels = runner.model.module.memory_bank.label_bank
            if new_labels.is_cuda:
                new_labels = new_labels.cpu()
            self.evaluate(runner, new_labels.numpy())

    def after_train_epoch(self, runner):
        if self.online_labels:  # online labels
            # save cluster
            if self.every_n_epochs(runner, self.interval) and runner.rank == 0:
                features = runner.model.module.memory_bank.feature_bank
                new_labels = runner.model.module.memory_bank.label_bank
                if new_labels.is_cuda:
                    new_labels = new_labels.cpu()
                np.save(
                    f'{runner.work_dir}/feature_epoch_{runner.epoch + 1}.npy',
                    features.cpu().numpy())
                np.save(
                    f'{runner.work_dir}/cluster_epoch_{runner.epoch + 1}.npy',
                    new_labels.numpy())
        else:  # offline labels
            if self.every_n_epochs(runner, self.interval):
                self.deepcluster(runner)

    def deepcluster(self, runner):
        # step 1: get features
        runner.model.eval()
        features = self.extractor(runner)
        runner.model.train()

        # step 2: get labels
        if not self.dist_mode or (self.dist_mode and runner.rank == 0):
            clustering_algo = _clustering.__dict__[self.clustering_type](
                **self.clustering_cfg)
            # Features are normalized during clustering
            clustering_algo.cluster(features, verbose=True)
            assert isinstance(clustering_algo.labels, np.ndarray)
            new_labels = clustering_algo.labels.astype(np.int64)
            if self.init_memory:
                np.save(f'{runner.work_dir}/cluster_epoch_{runner.epoch}.npy',
                        new_labels)
            else:
                np.save(
                    f'{runner.work_dir}/cluster_epoch_{runner.epoch + 1}.npy',
                    new_labels)
            self.evaluate(runner, new_labels)
        else:
            new_labels = np.zeros((len(self.data_loaders[0].dataset), ),
                                  dtype=np.int64)

        if self.dist_mode:
            new_labels_tensor = torch.from_numpy(new_labels).cuda()
            dist.broadcast(new_labels_tensor, 0)
            new_labels = new_labels_tensor.cpu().numpy()

        # step 3 (optional): assign offline labels
        if not (self.online_labels or self.init_memory):
            runner.model.module.memory_bank.assign_label(new_labels)

        # step 4 (before run): initialize memory
        if self.init_memory:
            runner.model.module.memory_bank.init_memory(features, new_labels)
            self.init_memory = False

    def evaluate(self, runner, new_labels):
        histogram = np.bincount(
            new_labels, minlength=runner.model.module.memory_bank.num_classes)
        empty_cls = (histogram == 0).sum()
        minimal_cls_size, maximal_cls_size = histogram.min(), histogram.max()
        if runner.rank == 0:
            print_log(
                f'empty_num: {empty_cls.item()}\t'
                f'min_cluster: {minimal_cls_size.item()}\t'
                f'max_cluster: {maximal_cls_size.item()}',
                logger='mmselfsup')



_base_ = [
    '../_base_/models/vit-base-p16_ft.py',
    '../_base_/datasets/imagenet.py',
    '../_base_/schedules/adamw_coslr-100e_in1k.py',
    '../_base_/default_runtime.py',
]
# maskfeat fine-tuning setting

# dataset
img_norm_cfg = dict(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
train_pipeline = [
    dict(
        type='RandomAug',
        input_size=224,
        color_jitter=0.4,
        auto_augment='rand-m9-mstd0.5-inc1',
        interpolation='bicubic',
        re_prob=0.25,
        re_mode='pixel',
        re_count=1,
        mean=(0.485, 0.456, 0.406),
        std=(0.229, 0.224, 0.225))
]
test_pipeline = [
    dict(type='Resize', size=256, interpolation=3),
    dict(type='CenterCrop', size=224),
    dict(type='ToTensor'),
    dict(type='Normalize', **img_norm_cfg)
]
data = dict(
    samples_per_gpu=256,
    drop_last=False,
    workers_per_gpu=32,
    train=dict(pipeline=train_pipeline),
    val=dict(pipeline=test_pipeline))

# model
model = dict(
    backbone=dict(init_cfg=dict()),
    head=dict(
        type='MaskFeatFinetuneHead',
        num_classes=1000,
        embed_dim=768,
        label_smooth_val=0.1))

# optimizer
optimizer = dict(
    lr=0.002 * 8 / 2,
    betas=(0.9, 0.999),
    weight_decay=0.05,
    paramwise_options={
        'ln': dict(weight_decay=0.),
        'bias': dict(weight_decay=0.),
        'pos_embed': dict(weight_decay=0.),
        'cls_token': dict(weight_decay=0.),
    },
    constructor='TransformerFinetuneConstructor',
    model_type='vit',
    layer_decay=0.65)

# learning policy
lr_config = dict(
    policy='CosineAnnealing',
    min_lr=1e-6,
    warmup='linear',
    warmup_iters=20,
    warmup_ratio=1e-08,
    warmup_by_epoch=True)

# runtime
checkpoint_config = dict(interval=1, max_keep_ckpts=3, out_dir='')
persistent_workers = True
log_config = dict(
    interval=100, hooks=[
        dict(type='TextLoggerHook'),
    ])



# dataset settings
data_source = 'ImageNet'
train_dataset_type = 'MultiViewDataset'
extract_dataset_type = 'SingleViewDataset'
img_norm_cfg = dict(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
train_pipeline = [
    dict(type='RandomResizedCrop', size=224, scale=(0.2, 1.)),
    dict(
        type='RandomAppliedTrans',
        transforms=[
            dict(
                type='ColorJitter',
                brightness=0.4,
                contrast=0.4,
                saturation=0.4,
                hue=0.1)
        ],
        p=0.8),
    dict(type='RandomGrayscale', p=0.2),
    dict(type='GaussianBlur', sigma_min=0.1, sigma_max=2.0, p=0.5),
    dict(type='RandomHorizontalFlip'),
]
extract_pipeline = [
    dict(type='Resize', size=256),
    dict(type='CenterCrop', size=224),
]

# prefetch
prefetch = False
if not prefetch:
    train_pipeline.extend(
        [dict(type='ToTensor'),
         dict(type='Normalize', **img_norm_cfg)])
    extract_pipeline.extend(
        [dict(type='ToTensor'),
         dict(type='Normalize', **img_norm_cfg)])

# dataset summary
data = dict(
    samples_per_gpu=32,  # total 32*8=256
    replace=True,
    workers_per_gpu=4,
    drop_last=True,
    train=dict(
        type=train_dataset_type,
        data_source=dict(
            type=data_source,
            data_prefix='data/imagenet/train',
            ann_file='data/imagenet/meta/train.txt',
        ),
        num_views=[2],
        pipelines=[train_pipeline],
        prefetch=prefetch))

# additional hooks
num_classes = 10000
custom_hooks = [
    dict(
        type='InterCLRHook',
        extractor=dict(
            samples_per_gpu=256,
            workers_per_gpu=8,
            dataset=dict(
                type=extract_dataset_type,
                data_source=dict(
                    type=data_source,
                    data_prefix='data/imagenet/train',
                    ann_file='data/imagenet/meta/train.txt',
                ),
                pipeline=extract_pipeline,
                prefetch=prefetch),
            prefetch=prefetch,
            img_norm_cfg=img_norm_cfg),
        clustering=dict(type='Kmeans', k=num_classes, pca_dim=-1),  # no pca
        centroids_update_interval=10,  # iter
        deal_with_small_clusters_interval=1,
        evaluate_interval=50,
        warmup_epochs=0,
        init_memory=True,
        initial=True,  # call initially
        online_labels=True,
        interval=10)  # same as the checkpoint interval
]



# model settings
model = dict(
    type='InterCLRMoCo',
    queue_len=65536,
    feat_dim=128,
    momentum=0.999,
    backbone=dict(
        type='ResNet',
        depth=50,
        in_channels=3,
        out_indices=[4],  # 0: conv-1, x: stage-x
        norm_cfg=dict(type='BN')),
    neck=dict(
        type='MoCoV2Neck',
        in_channels=2048,
        hid_channels=2048,
        out_channels=128,
        with_avg_pool=True),
    head=dict(type='ContrastiveHead', temperature=0.2),
    memory_bank=dict(
        type='InterCLRMemory',
        length=1281167,
        feat_dim=128,
        momentum=1.,
        num_classes=10000,
        min_cluster=20,
        debug=False),
    online_labels=True,
    neg_num=16384,
    neg_sampling='semihard',  # 'hard', 'semihard', 'random', 'semieasy'
    semihard_neg_pool_num=128000,
    semieasy_neg_pool_num=128000,
    intra_cos_marign_loss=False,
    intra_cos_margin=0,
    inter_cos_marign_loss=True,
    inter_cos_margin=-0.5,
    intra_loss_weight=0.75,
    inter_loss_weight=0.25,
    share_neck=True,
    num_classes=10000)



# dataset settings
data_source = 'ImageNet'
dataset_type = 'MultiViewDataset'
num_crops = [2, 6]
color_distort_strength = 1.0
img_norm_cfg = dict(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
train_pipeline1 = [
    dict(type='RandomResizedCrop', size=224, scale=(0.14, 1.)),
    dict(
        type='RandomAppliedTrans',
        transforms=[
            dict(
                type='ColorJitter',
                brightness=0.8 * color_distort_strength,
                contrast=0.8 * color_distort_strength,
                saturation=0.8 * color_distort_strength,
                hue=0.2 * color_distort_strength)
        ],
        p=0.8),
    dict(type='RandomGrayscale', p=0.2),
    dict(type='GaussianBlur', sigma_min=0.1, sigma_max=2.0, p=0.5),
    dict(type='RandomHorizontalFlip', p=0.5),
]
train_pipeline2 = [
    dict(type='RandomResizedCrop', size=96, scale=(0.05, 0.14)),
    dict(
        type='RandomAppliedTrans',
        transforms=[
            dict(
                type='ColorJitter',
                brightness=0.8 * color_distort_strength,
                contrast=0.8 * color_distort_strength,
                saturation=0.8 * color_distort_strength,
                hue=0.2 * color_distort_strength)
        ],
        p=0.8),
    dict(type='RandomGrayscale', p=0.2),
    dict(type='GaussianBlur', sigma_min=0.1, sigma_max=2.0, p=0.5),
    dict(type='RandomHorizontalFlip', p=0.5),
]

# prefetch
prefetch = False
if not prefetch:
    train_pipeline1.extend(
        [dict(type='ToTensor'),
         dict(type='Normalize', **img_norm_cfg)])
    train_pipeline2.extend(
        [dict(type='ToTensor'),
         dict(type='Normalize', **img_norm_cfg)])

# dataset summary
data = dict(
    samples_per_gpu=32,  # total 32*8=256
    workers_per_gpu=4,
    drop_last=True,
    train=dict(
        type=dataset_type,
        data_source=dict(
            type=data_source,
            data_prefix='data/imagenet/train',
            ann_file='data/imagenet/meta/train.txt',
        ),
        num_views=num_crops,
        pipelines=[train_pipeline1, train_pipeline2],
        prefetch=prefetch))



# Copyright (c) OpenMMLab. All rights reserved.
import math

import numpy as np
import torch
import torch.nn as nn

from mmselfsup.utils import (batch_shuffle_ddp, batch_unshuffle_ddp,
                             concat_all_gather)
from ..builder import (ALGORITHMS, build_backbone, build_head, build_memory,
                       build_neck)
from .base import BaseModel


@ALGORITHMS.register_module()
class InterCLRMoCo(BaseModel):
    """MoCo-InterCLR.

    Official implementation of `Delving into Inter-Image Invariance for
    Unsupervised Visual Representations <https://arxiv.org/abs/2008.11702>`_.
    The clustering operation is in `core/hooks/interclr_hook.py`.

    Args:
        backbone (dict): Config dict for module of backbone.
        neck (dict): Config dict for module of deep features to compact feature
            vectors. Defaults to None.
        head (dict): Config dict for module of loss functions.
            Defaults to None.
        queue_len (int): Number of negative keys maintained in the queue.
            Defaults to 65536.
        feat_dim (int): Dimension of compact feature vectors. Defaults to 128.
        momentum (float): Momentum coefficient for the momentum-updated
            encoder. Defaults to 0.999.
        memory_bank (dict): Config dict for module of memory banks.
            Defaults to None.
        online_labels (bool): Whether to use online labels. Defaults to True.
        neg_num (int): Number of negative samples for inter-image branch.
            Defaults to 16384.
        neg_sampling (str): Negative sampling strategy. Support 'hard',
            'semihard', 'random', 'semieasy'. Defaults to 'semihard'.
        semihard_neg_pool_num (int): Number of negative samples for semi-hard
            nearest neighbor pool. Defaults to 128000.
        semieasy_neg_pool_num (int): Number of negative samples for semi-easy
            nearest neighbor pool. Defaults to 128000.
        intra_cos_marign_loss (bool): Whether to use a cosine margin for
            intra-image branch. Defaults to False.
        intra_cos_marign (float): Intra-image cosine margin. Defaults to 0.
        intra_arc_marign_loss (bool): Whether to use an arc margin for
            intra-image branch. Defaults to False.
        intra_arc_marign (float): Intra-image arc margin. Defaults to 0.
        inter_cos_marign_loss (bool): Whether to use a cosine margin for
            inter-image branch. Defaults to True.
        inter_cos_marign (float): Inter-image cosine margin. Defaults to -0.5.
        inter_arc_marign_loss (bool): Whether to use an arc margin for
            inter-image branch. Defaults to False.
        inter_arc_marign (float): Inter-image arc margin. Defaults to 0.
        intra_loss_weight (float): Loss weight for intra-image branch.
            Defaults to 0.75.
        inter_loss_weight (float): Loss weight for inter-image branch.
            Defaults to 0.25.
        share_neck (bool): Whether to share the neck for intra- and inter-image
            branches. Defaults to True.
        num_classes (int): Number of clusters. Defaults to 10000.
    """

    def __init__(self,
                 backbone,
                 neck=None,
                 head=None,
                 queue_len=65536,
                 feat_dim=128,
                 momentum=0.999,
                 memory_bank=None,
                 online_labels=True,
                 neg_num=16384,
                 neg_sampling='semihard',
                 semihard_neg_pool_num=128000,
                 semieasy_neg_pool_num=128000,
                 intra_cos_marign_loss=False,
                 intra_cos_margin=0,
                 intra_arc_marign_loss=False,
                 intra_arc_margin=0,
                 inter_cos_marign_loss=True,
                 inter_cos_margin=-0.5,
                 inter_arc_marign_loss=False,
                 inter_arc_margin=0,
                 intra_loss_weight=0.75,
                 inter_loss_weight=0.25,
                 share_neck=True,
                 num_classes=10000,
                 init_cfg=None,
                 **kwargs):
        super(InterCLRMoCo, self).__init__(init_cfg)
        self.encoder_q = nn.Sequential(
            build_backbone(backbone), build_neck(neck))
        self.encoder_k = nn.Sequential(
            build_backbone(backbone), build_neck(neck))
        if not share_neck:
            self.inter_neck_q = build_neck(neck)
            self.inter_neck_k = build_neck(neck)
        self.backbone = self.encoder_q[0]
        self.neck = self.encoder_q[1]
        self.head = build_head(head)
        self.memory_bank = build_memory(memory_bank)

        # moco params
        self.queue_len = queue_len
        self.momentum = momentum

        # interclr params
        self.online_labels = online_labels
        self.neg_num = neg_num
        self.neg_sampling = neg_sampling
        self.semihard_neg_pool_num = semihard_neg_pool_num
        self.semieasy_neg_pool_num = semieasy_neg_pool_num
        self.intra_cos = intra_cos_marign_loss
        self.intra_cos_margin = intra_cos_margin
        self.intra_arc = intra_arc_marign_loss
        self.intra_arc_margin = intra_arc_margin
        self.inter_cos = inter_cos_marign_loss
        self.inter_cos_margin = inter_cos_margin
        self.inter_arc = inter_arc_marign_loss
        self.inter_arc_margin = inter_arc_margin
        self.intra_loss_weight = intra_loss_weight
        self.inter_loss_weight = inter_loss_weight
        self.share_neck = share_neck
        self.num_classes = num_classes

        # create the queue
        self.register_buffer('queue', torch.randn(feat_dim, queue_len))
        self.queue = nn.functional.normalize(self.queue, dim=0)
        self.register_buffer('queue_ptr', torch.zeros(1, dtype=torch.long))

    def init_weights(self):
        """Initialize base_encoder with init_cfg defined in backbone."""
        super(InterCLRMoCo, self).init_weights()

        for param_q, param_k in zip(self.encoder_q.parameters(),
                                    self.encoder_k.parameters()):
            param_k.data.copy_(param_q.data)
            param_k.requires_grad = False
        if not self.share_neck:
            for param_q, param_k in zip(self.inter_neck_q.parameters(),
                                        self.inter_neck_k.parameters()):
                param_k.data.copy_(param_q.data)
                param_k.requires_grad = False

    @torch.no_grad()
    def _momentum_update_key_encoder(self):
        """Momentum update of the key encoder."""
        for param_q, param_k in zip(self.encoder_q.parameters(),
                                    self.encoder_k.parameters()):
            param_k.data = param_k.data * self.momentum + \
                           param_q.data * (1. - self.momentum)
        if not self.share_neck:
            for param_q, param_k in zip(self.inter_neck_q.parameters(),
                                        self.inter_neck_k.parameters()):
                param_k.data = param_k.data * self.momentum + \
                               param_q.data * (1. - self.momentum)

    @torch.no_grad()
    def _dequeue_and_enqueue(self, keys):
        """Update queue."""
        # normalize
        keys = nn.functional.normalize(keys, dim=1)
        # gather keys before updating queue
        keys = concat_all_gather(keys)

        batch_size = keys.shape[0]

        ptr = int(self.queue_ptr)
        assert self.queue_len % batch_size == 0  # for simplicity

        # replace the keys at ptr (dequeue and enqueue)
        self.queue[:, ptr:ptr + batch_size] = keys.transpose(0, 1)
        ptr = (ptr + batch_size) % self.queue_len  # move pointer

        self.queue_ptr[0] = ptr

    def contrast_intra(self, q, k):
        """Intra-image invariance learning.

        Args:
            q (Tensor): Query features with shape (N, C).
            k (Tensor): Key features with shape (N, C).

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        # normalize
        q = nn.functional.normalize(q, dim=1)
        k = nn.functional.normalize(k, dim=1)
        # compute logits
        # Einstein sum is more intuitive
        # positive logits: Nx1
        pos_logits = torch.einsum('nc,nc->n', [q, k]).unsqueeze(-1)
        # negative logits: NxK
        neg_logits = torch.einsum('nc,ck->nk',
                                  [q, self.queue.clone().detach()])

        # use cosine margin
        if self.intra_cos:
            cosine = pos_logits.clone()
            phi = cosine - self.intra_cos_margin
            pos_logits.copy_(phi)
        # use arc margin
        if self.intra_arc:
            cosine = pos_logits.clone()
            sine = torch.sqrt((1.0 - cosine**2).clamp(0, 1))
            phi = cosine * math.cos(self.intra_arc_margin) - sine * math.sin(
                self.intra_arc_margin)
            if self.intra_arc_margin < 0:
                phi = torch.where(
                    cosine < math.cos(self.intra_arc_margin), phi, cosine +
                    math.sin(self.intra_arc_margin) * self.intra_arc_margin)
            else:
                phi = torch.where(
                    cosine > math.cos(math.pi - self.intra_arc_margin), phi,
                    cosine - math.sin(math.pi - self.intra_arc_margin) *
                    self.intra_arc_margin)
            pos_logits.copy_(phi)

        losses = self.head(pos_logits, neg_logits)

        return losses

    def contrast_inter(self, q, idx):
        """Inter-image invariance learning.

        Args:
            q (Tensor): Query features with shape (N, C).
            idx (Tensor): Index corresponding to each query.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        # normalize
        feat_norm = nn.functional.normalize(q, dim=1)
        bs, feat_dim = feat_norm.shape[:2]
        # positive sampling
        pos_label = self.memory_bank.label_bank[idx.cpu()]
        pos_idx_list = []
        for i, l in enumerate(pos_label):
            pos_idx_pool = torch.where(
                self.memory_bank.label_bank == l)[0]  # positive index pool
            pos_i = torch.zeros(
                1, dtype=torch.long).random_(0, pos_idx_pool.size(0))
            pos_idx_list.append(pos_idx_pool[pos_i])
        pos_idx = torch.cuda.LongTensor(pos_idx_list)
        # negative sampling
        if self.neg_sampling == 'random':  # random negative sampling
            pos_label = pos_label.cuda().unsqueeze(1)
            neg_idx = self.memory_bank.multinomial.draw(
                bs * self.neg_num).view(bs, -1)
            while True:
                neg_label = self.memory_bank.label_bank[neg_idx.cpu()].cuda()
                pos_in_neg_idx = (neg_label == pos_label)
                if pos_in_neg_idx.sum().item() > 0:
                    neg_idx[
                        pos_in_neg_idx] = self.memory_bank.multinomial.draw(
                            pos_in_neg_idx.sum().item())
                else:
                    break
        elif self.neg_sampling == 'semihard':  # semihard negative sampling
            pos_label = pos_label.cuda().unsqueeze(1)
            similarity = torch.mm(feat_norm.detach(),
                                  self.memory_bank.feature_bank.permute(1, 0))
            _, neg_I = torch.topk(
                similarity, self.semihard_neg_pool_num, dim=1, sorted=False)
            weights = torch.ones((bs, self.semihard_neg_pool_num),
                                 dtype=torch.float,
                                 device='cuda')
            neg_i = torch.multinomial(weights, self.neg_num)
            neg_idx = torch.gather(neg_I, 1, neg_i)
            while True:
                neg_label = self.memory_bank.label_bank[neg_idx.cpu()].cuda()
                pos_in_neg_idx = (neg_label == pos_label)
                if pos_in_neg_idx.sum().item() > 0:
                    neg_i = torch.multinomial(weights, self.neg_num)
                    neg_idx[pos_in_neg_idx] = torch.gather(
                        neg_I, 1, neg_i)[pos_in_neg_idx]
                else:
                    break
        elif self.neg_sampling == 'semieasy':  # semieasy negative sampling
            pos_label = pos_label.cuda().unsqueeze(1)
            similarity = torch.mm(feat_norm.detach(),
                                  self.memory_bank.feature_bank.permute(1, 0))
            _, neg_I = torch.topk(
                similarity,
                self.semieasy_neg_pool_num,
                dim=1,
                largest=False,
                sorted=False)
            weights = torch.ones((bs, self.semieasy_neg_pool_num),
                                 dtype=torch.float,
                                 device='cuda')
            neg_i = torch.multinomial(weights, self.neg_num)
            neg_idx = torch.gather(neg_I, 1, neg_i)
            while True:
                neg_label = self.memory_bank.label_bank[neg_idx.cpu()].cuda()
                pos_in_neg_idx = (neg_label == pos_label)
                if pos_in_neg_idx.sum().item() > 0:
                    neg_i = torch.multinomial(weights, self.neg_num)
                    neg_idx[pos_in_neg_idx] = torch.gather(
                        neg_I, 1, neg_i)[pos_in_neg_idx]
                else:
                    break
        elif self.neg_sampling == 'hard':  # hard negative sampling
            similarity = torch.mm(feat_norm.detach(),
                                  self.memory_bank.feature_bank.permute(1, 0))
            maximal_cls_size = np.bincount(
                self.memory_bank.label_bank.numpy(),
                minlength=self.num_classes).max().item()
            _, neg_I = torch.topk(
                similarity, self.neg_num + maximal_cls_size, dim=1)
            neg_I = neg_I.cpu()
            neg_label = self.memory_bank.label_bank[neg_I].numpy()
            neg_idx_list = []
            for i, l in enumerate(pos_label):
                pos_in_neg_idx = np.where(neg_label[i] == l)[0]
                if len(pos_in_neg_idx) > 0:
                    neg_idx_pool = torch.from_numpy(
                        np.delete(neg_I[i].numpy(), pos_in_neg_idx))
                else:
                    neg_idx_pool = neg_I[i]
                neg_idx_list.append(neg_idx_pool[:self.neg_num])
            neg_idx = torch.stack(neg_idx_list, dim=0).cuda()
        else:
            raise Exception(
                f'No {self.neg_sampling} negative sampling strategy.')

        pos_feat = torch.index_select(self.memory_bank.feature_bank, 0,
                                      pos_idx)  # BXC
        neg_feat = torch.index_select(self.memory_bank.feature_bank, 0,
                                      neg_idx.flatten()).view(
                                          bs, self.neg_num, feat_dim)  # BxKxC

        pos_logits = torch.einsum('nc,nc->n',
                                  [pos_feat, feat_norm]).unsqueeze(-1)
        neg_logits = torch.bmm(neg_feat, feat_norm.unsqueeze(2)).squeeze(2)

        # use cosine margin
        if self.inter_cos:
            cosine = pos_logits.clone()
            phi = cosine - self.inter_cos_margin
            pos_logits.copy_(phi)
        # use arc margin
        if self.inter_arc:
            cosine = pos_logits.clone()
            sine = torch.sqrt((1.0 - cosine**2).clamp(0, 1))
            phi = cosine * math.cos(self.inter_arc_margin) - sine * math.sin(
                self.inter_arc_margin)
            if self.inter_arc_margin < 0:
                phi = torch.where(
                    cosine < math.cos(self.inter_arc_margin), phi, cosine +
                    math.sin(self.inter_arc_margin) * self.inter_arc_margin)
            else:
                phi = torch.where(
                    cosine > math.cos(math.pi - self.inter_arc_margin), phi,
                    cosine - math.sin(math.pi - self.inter_arc_margin) *
                    self.inter_arc_margin)
            pos_logits.copy_(phi)

        losses = self.head(pos_logits, neg_logits)

        return losses

    def extract_feat(self, img):
        """Function to extract features from backbone.

        Args:
            img (Tensor): Input images of shape (N, C, H, W).
                Typically these should be mean centered and std scaled.

        Returns:
            tuple[Tensor]: backbone outputs.
        """
        x = self.backbone(img)
        return x

    def forward_train(self, img, idx, **kwargs):
        """Forward computation during training.

        Args:
            img (list[Tensor]): A list of input images with shape
                (N, C, H, W). Typically these should be mean centered
                and std scaled.
            idx (Tensor): Index corresponding to each image.
            kwargs: Any keyword arguments to be used to forward.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        assert isinstance(img, list)
        im_q = img[0]
        im_k = img[1]

        # compute query features
        q_b = self.encoder_q[0](im_q)  # backbone features
        q = self.encoder_q[1](q_b)[0]  # queries: NxC
        if not self.share_neck:
            q2 = self.inter_neck_q(q_b)[0]  # inter queries: NxC

        # compute key features
        with torch.no_grad():  # no gradient to keys
            self._momentum_update_key_encoder()  # update the key encoder

            # shuffle for making use of BN
            im_k, idx_unshuffle = batch_shuffle_ddp(im_k)

            k_b = self.encoder_k[0](im_k)  # backbone features
            k = self.encoder_k[1](k_b)[0]  # keys: NxC
            if not self.share_neck:
                k2 = self.inter_neck_k(k_b)[0]  # inter keys: NxC

            # undo shuffle
            k = batch_unshuffle_ddp(k, idx_unshuffle)
            if not self.share_neck:
                k2 = batch_unshuffle_ddp(k2, idx_unshuffle)

        idx = idx.cuda()
        self.memory_bank.broadcast_feature_bank()
        # compute intra loss
        intra_losses = self.contrast_intra(q, k)
        # compute inter loss
        if self.share_neck:
            inter_losses = self.contrast_inter(q, idx)
        else:
            inter_losses = self.contrast_inter(q2, idx)
        losses = dict()
        losses['intra_loss'] = self.intra_loss_weight * intra_losses['loss']
        losses['inter_loss'] = self.inter_loss_weight * inter_losses['loss']

        self._dequeue_and_enqueue(k)

        # update memory bank
        if self.online_labels:
            if self.share_neck:
                change_ratio = self.memory_bank.update_samples_memory(
                    idx, k.detach())
            else:
                change_ratio = self.memory_bank.update_samples_memory(
                    idx, k2.detach())
            losses['change_ratio'] = change_ratio
        else:
            if self.share_neck:
                self.memory_bank.update_simple_memory(idx, k.detach())
            else:
                self.memory_bank.update_simple_memory(idx, k2.detach())

        return losses



_base_ = [
    '../_base_/models/fcn_r50-d8.py',
    '../_base_/datasets/cityscapes_769x769.py', '../_base_/default_runtime.py',
    '../_base_/schedules/schedule_40k.py'
]
model = dict(
    decode_head=dict(align_corners=True),
    auxiliary_head=dict(align_corners=True),
    test_cfg=dict(mode='slide', crop_size=(769, 769), stride=(513, 513)))
